import torch
from GANDLF.grad_clipping.clip_gradients import dispatch_clip_grad_


class GradScaler:
    def __init__(self):
        """The initialization of the GradScaler class."""
        self._scaler = torch.cuda.amp.GradScaler()

    def __call__(
        self,
        loss,
        optimizer,
        clip_grad=None,
        clip_mode="norm",
        parameters=None,
        create_graph=False,
    ):
        self._scaler.scale(loss).backward(create_graph=create_graph)
        if clip_grad is not None:
            assert parameters is not None
            # unscale the gradients of optimizer's assigned params in-place
            self._scaler.unscale_(optimizer)
            if (clip_mode is None) or (str(clip_mode).lower() == "none"):
                clip_mode = "norm"  # default, in case none gets passed
            dispatch_clip_grad_(parameters, clip_grad, mode=clip_mode)
        self._scaler.step(optimizer)
        self._scaler.update()

    def state_dict(self):
        return self._scaler.state_dict()

    def load_state_dict(self, state_dict):
        self._scaler.load_state_dict(state_dict)


def model_parameters_exclude_head(model, clip_mode=None):
    exclude_head = False
    if clip_mode is not None:
        if clip_mode == "agc":
            exclude_head = True
    if exclude_head:
        return [p for p in model.parameters()][:-2]
    else:
        return model.parameters()



import torch
import torch.nn.functional as F
from skimage.measure import label
import numpy as np
from scipy.ndimage import binary_fill_holes, binary_closing
from GANDLF.utils.generic import get_array_from_image_or_tensor


def torch_morphological(input_image, kernel_size=1, mode="dilation"):
    """
    This function enables morphological operations using torch. Adapted from https://github.com/DIVA-DIA/Generating-Synthetic-Handwritten-Historical-Documents/blob/e6a798dc2b374f338804222747c56cb44869af5b/HTR_ctc/utils/auxilary_functions.py#L10.

    Args:
        input_image (torch.Tensor): The input image.
        kernel_size (list): The size of the window to take a max over.
        mode (str): The type of morphological operation to perform.

    Returns:
        torch.Tensor: The output image after morphological operations.
    """

    if len(input_image.shape) == 4:
        max_pool = F.max_pool2d
    elif len(input_image.shape) == 5:
        max_pool = F.max_pool3d
    else:
        raise ValueError("Input image has invalid shape for morphological operations.")

    if mode == "dilation":
        output_image = max_pool(
            input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "erosion":
        output_image = -max_pool(
            -input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "closing":
        output_image = max_pool(
            input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
        output_image = -max_pool(
            -output_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "opening":
        output_image = -max_pool(
            -input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
        output_image = max_pool(
            output_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )

    return output_image


def fill_holes(input_image, params=None):
    """
    This function fills holes in masks.

    Args:
        input_image (torch.Tensor): The input image.
        params (dict): The parameters dict; unused.

    Returns:
        torch.Tensor: The output image after morphological operations.
    """
    input_image_array = get_array_from_image_or_tensor(input_image).astype(int)
    input_image_array_closed = binary_closing(input_image_array)
    # Fill the holes in binary objects
    output_array = binary_fill_holes(input_image_array_closed).astype(int)

    return torch.from_numpy(output_array)


def cca(input_image):
    """
    This function performs connected component analysis on the input image.

    Args:
        input_image (torch.Tensor): The input image.
        params (dict): The parameters dict;

    Returns:
        torch.Tensor: The output image after morphological operations.
    """
    seg = get_array_from_image_or_tensor(input_image)
    mask = seg != 0

    connectivity = input_image.dim() - 1
    labels_connected = label(mask, connectivity=connectivity)
    labels_connected_sizes = [
        np.sum(labels_connected == i) for i in np.unique(labels_connected)
    ]
    largest_region = np.argmax(labels_connected_sizes[1:]) + 1
    seg[labels_connected != largest_region] = 0
    return seg



# -*- coding: utf-8 -*-
"""Add Conv Block for MSDNet"""


def add_conv_block(
    Conv, BatchNorm, in_ch=1, out_ch=1, kernel_size=3, dilate=1, last=False
):
    """
    Helper function
    """
    pad = dilate if not last else 0
    conv_1 = Conv(in_ch, out_ch, kernel_size, padding=pad, dilation=dilate)
    bn_1 = BatchNorm(out_ch)

    return [conv_1, bn_1]



import torch.nn as nn
import torch.nn.functional as F


class IncDownsamplingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        Conv,
        Dropout,
        InstanceNorm,
        leakiness=1e-2,
        kernel_size=1,
        conv_bias=True,
        inst_norm_affine=True,
        lrelu_inplace=True,
    ):
        nn.Module.__init__(self)
        self.output_channels = output_channels
        self.input_channels = input_channels
        self.leakiness = leakiness
        self.conv_bias = conv_bias
        self.lrelu_inplace = lrelu_inplace
        self.inst_norm_affine = inst_norm_affine
        self.inst_norm = InstanceNorm(
            output_channels, affine=self.inst_norm_affine, track_running_stats=True
        )
        self.down = Conv(
            input_channels,
            output_channels,
            kernel_size=1,
            stride=2,
            padding=0,
            bias=self.conv_bias,
        )

    def forward(self, x):
        x = F.leaky_relu(
            self.inst_norm(self.down(x)),
            negative_slope=self.leakiness,
            inplace=self.lrelu_inplace,
        )
        return x



from .segmentation import MCD_loss
from .regression import CCE_Generic, CE, CE_Logits


def DCCE(pm, gt, params):
    dcce_loss = MCD_loss(pm, gt, params) + CCE_Generic(pm, gt, params, CE)
    return dcce_loss


def DCCE_Logits(pm, gt, params):
    dcce_loss = MCD_loss(pm, gt, params) + CCE_Generic(pm, gt, params, CE_Logits)
    return dcce_loss



import torch


# Dice scores and dice losses
def dice(output, label):
    """
    This function computes a dice score between two tensors

    Parameters
    ----------
    output : Tensor
        Output predicted generally by the network
    label : Tensor
        Required target label to match the output with

    Returns
    -------
    Tensor
        Computed Dice Score

    """
    smooth = 1e-7
    iflat = output.contiguous().view(-1)
    tflat = label.contiguous().view(-1)
    intersection = (iflat * tflat).sum()
    return (2.0 * intersection + smooth) / (iflat.sum() + tflat.sum() + smooth)


def MCD(pm, gt, num_class, weights=None, ignore_class=None, loss_type=0):
    """
    These weights should be the dice weights, not dice weights
    loss_type:
        0: no loss, normal dice calculation
        1: dice loss, (1-dice)
        2: log dice, -log(dice)
    """
    acc_dice = 0
    for i in range(0, num_class):  # 0 is background
        currentDice = dice(gt[:, i, ...], pm[:, i, ...])
        if loss_type == 1:
            currentDice = 1 - currentDice  # subtract from 1 because this is a loss
        elif loss_type == 2:
            # negative because we want positive losses
            currentDice = -torch.log(currentDice + torch.finfo(torch.float32).eps)
        if weights is not None:
            currentDice = currentDice * weights[i]
        acc_dice += currentDice
    if weights is None:
        acc_dice /= num_class  # we should not be considering 0
    return acc_dice


def MCD_loss(pm, gt, params):
    """
    These weights should be the penalty weights, not dice weights
    """
    return MCD(pm, gt, len(params["model"]["class_list"]), params["weights"], None, 1)


def MCD_log_loss(pm, gt, params):
    """
    These weights should be the penalty weights, not dice weights
    """
    return MCD(pm, gt, len(params["model"]["class_list"]), params["weights"], None, 2)


def tversky(inp, target, alpha):
    smooth = 1e-7
    iflat = inp.view(-1)
    tflat = target.view(-1)
    intersection = (iflat * tflat).sum()
    fps = (iflat * (1 - tflat)).sum()
    fns = ((1 - iflat) * tflat).sum()
    denominator = intersection + (alpha * fps) + ((1 - alpha) * fns) + smooth
    return (intersection + smooth) / denominator


def tversky_loss(inp, target, alpha=1):
    tversky_val = tversky(inp, target, alpha)
    return 1 - tversky_val


def MCT_loss(inp, target, params):
    acc_tv_loss = 0
    for i in range(0, len(params["model"]["class_list"])):
        curr_loss = tversky_loss(inp[:, i, ...], target[:, i, ...])
        if params["weights"] is not None:
            curr_loss = curr_loss * params["weights"][i]
        acc_tv_loss += curr_loss
    if params["weights"] is None:
        acc_tv_loss /= len(params["model"]["class_list"])
    return acc_tv_loss


def KullbackLeiblerDivergence(mu, logvar, params=None):
    loss = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp(), dim=-1)
    return loss.mean()



import os, pickle
from pathlib import Path

from GANDLF.training_manager import TrainingManager, TrainingManager_split
from GANDLF.inference_manager import InferenceManager
from GANDLF.parseConfig import parseConfig
from GANDLF.utils import populate_header_in_parameters, parseTrainingCSV, parseTestingCSV

def main_run(
    data_csv, config_file, model_dir, train_mode, device, resume, reset, output_dir=None
):
    """
    Main function that runs the training and inference.

    Args:
        data_csv (str): The CSV file of the training data.
        config_file (str): The YAML file of the training configuration.
        model_dir (str): The model directory; for training, model is written out here, and for inference, trained model is expected here.
        train_mode (bool): Whether to train or infer.
        device (str): The device type.
        resume (bool): Whether the previous run will be resumed or not.
        reset (bool): Whether the previous run will be reset or not.
        output_dir (str): The output directory for the inference session.

    Returns:
        None
    """
    file_data_full = data_csv
    model_parameters = config_file
    device = device
    parameters = parseConfig(model_parameters)
    parameters["device_id"] = -1
    # in case the data being passed is already processed, check if the previous parameters exists,
    # and if it does, compare the two and if they are the same, ensure no preprocess is done.
    model_parameters_prev = os.path.join(os.path.dirname(model_dir), "parameters.pkl")
    if train_mode:
        if not (reset) or not (resume):
            print(
                "Trying to resume training without changing any parameters from previous run.",
                flush=True,
            )
            if os.path.exists(model_parameters_prev):
                parameters_prev = pickle.load(open(model_parameters_prev, "rb"))
                assert (
                    parameters == parameters_prev
                ), "The parameters are not the same as the ones stored in the previous run, please re-check."
        parameters["output_dir"] = model_dir
        Path(parameters["output_dir"]).mkdir(parents=True, exist_ok=True)

    if "-1" in device:
        device = "cpu"

    # parse training CSV
    if "," in file_data_full:
        # training and validation pre-split
        data_full = None
        all_csvs = file_data_full.split(",")
        data_train, headers_train = parseTrainingCSV(all_csvs[0], train=train_mode)
        data_validation, headers_validation = parseTrainingCSV(
            all_csvs[1], train=train_mode
        )
        assert (
            headers_train == headers_validation
        ), "The training and validation CSVs do not have the same header information."

        # testing data is present
        data_testing = None
        headers_testing = headers_train
        if len(all_csvs) == 3:
            data_testing, headers_testing = parseTrainingCSV(
                all_csvs[2], train=train_mode
            )
        assert (
            headers_train == headers_testing
        ), "The training and testing CSVs do not have the same header information."

        parameters = populate_header_in_parameters(parameters, headers_train)
        # if we are here, it is assumed that the user wants to do training
        if train_mode:
            TrainingManager_split(
                dataframe_train=data_train,
                dataframe_validation=data_validation,
                dataframe_testing=data_testing,
                outputDir=parameters["output_dir"],
                parameters=parameters,
                device=device,
                resume=resume,
                reset=reset,
            )
    else:
        data_full, headers = parseTrainingCSV(file_data_full, train=train_mode)
        parameters = populate_header_in_parameters(parameters, headers)
        if train_mode:
            TrainingManager(
                dataframe=data_full,
                outputDir=parameters["output_dir"],
                parameters=parameters,
                device=device,
                resume=resume,
                reset=reset,
            )
        else:
            data_full, headers = parseTestingCSV(
                file_data_full, parameters["output_dir"]
            )
            InferenceManager(
                dataframe=data_full,
                modelDir=model_dir,
                outputDir=output_dir,
                parameters=parameters,
                device=device,
            )



import torch
from torch.nn import MSELoss, CrossEntropyLoss, L1Loss
from GANDLF.utils import one_hot


def CEL(out, target, params):
    if len(target.shape) > 1 and target.shape[-1] == 1:
        target = torch.squeeze(target, -1)

    weights = None
    if params["weights"]:
        weights = torch.FloatTensor(list(params["weights"].values()))
        weights = weights.float().to(target.device)

    cel = CrossEntropyLoss(weight=weights)
    return cel(out, target)


def CE_Logits(out, target):
    iflat = out.contiguous().view(-1)
    tflat = target.contiguous().view(-1)
    loss = torch.nn.BCEWithLogitsLoss()
    loss_val = loss(iflat, tflat)
    return loss_val


def CE(out, target):
    iflat = out.contiguous().view(-1)
    tflat = target.contiguous().view(-1)
    loss = torch.nn.BCELoss()
    loss_val = loss(iflat.float(), tflat.float())
    return loss_val


def CCE_Generic(out, target, params, CCE_Type):
    """
    Generic function to calculate CCE loss

    Args:
        out (torch.tensor): The predicted output value for each pixel. dimension: [batch, class, x, y, z].
        target (torch.tensor): The ground truth label for each pixel. dimension: [batch, class, x, y, z] factorial_class_list.
        params (dict): The parameter dictionary.
        CCE_Type (torch.nn): The CE loss function type.

    Returns:
        torch.tensor: The final loss value after taking multiple classes into consideration
    """

    acc_ce_loss = 0
    target = one_hot(target, params["model"]["class_list"]).type(out.dtype)
    for i in range(0, len(params["model"]["class_list"])):
        curr_ce_loss = CCE_Type(out[:, i, ...], target[:, i, ...])
        if params["weights"] is not None:
            curr_ce_loss = curr_ce_loss * params["weights"][i]
        acc_ce_loss += curr_ce_loss
    if params["weights"] is None:
        acc_ce_loss /= len(params["model"]["class_list"])
    return acc_ce_loss


def L1(output, label, reduction="mean", scaling_factor=1):
    """
    Calculate the mean square error between the output variable from the network and the target

    Parameters
    ----------
    output : torch.Tensor
        The output generated usually by the network
    target : torch.Tensor
        The label for the corresponding Tensor for which the output was generated
    reduction : string, optional
        DESCRIPTION. The default is 'mean'.
    scaling_factor : integer, optional
        The scaling factor to multiply the label with

    Returns
    -------
    loss : torch.Tensor
        Computed Mean Squared Error loss for the output and label

    """
    scaling_factor = torch.as_tensor(scaling_factor)
    label = label.float()
    label = label * scaling_factor
    loss_fn = L1Loss(reduction=reduction)
    iflat = output.contiguous().view(-1)
    tflat = label.contiguous().view(-1)
    loss = loss_fn(iflat, tflat)
    return loss


def L1_loss(inp, target, params):
    acc_mse_loss = 0
    # if inp.shape != target.shape:
    #     sys.exit('Input and target shapes are inconsistent')

    if inp.shape[0] == 1:
        if params is not None:
            acc_mse_loss += L1(
                inp,
                target,
                reduction=params["loss_function"]["l1"]["reduction"],
                scaling_factor=params["scaling_factor"],
            )
        else:
            acc_mse_loss += L1(inp, target)
        # for i in range(0, params["model"]["num_classes"]):
        #    acc_mse_loss += MSE(inp[i], target[i], reduction=params["loss_function"]['mse']["reduction"])
    else:
        if params is not None:
            for i in range(0, params["model"]["num_classes"]):
                acc_mse_loss += L1(
                    inp[:, i, ...],
                    target[:, i, ...],
                    reduction=params["loss_function"]["mse"]["reduction"],
                    scaling_factor=params["scaling_factor"],
                )
        else:
            for i in range(0, inp.shape[1]):
                acc_mse_loss += L1(inp[:, i, ...], target[:, i, ...])
    if params is not None:
        acc_mse_loss /= params["model"]["num_classes"]
    else:
        acc_mse_loss /= inp.shape[1]

    return acc_mse_loss


def MSE(output, label, reduction="mean", scaling_factor=1):
    """
    Calculate the mean square error between the output variable from the network and the target

    Parameters
    ----------
    output : torch.Tensor
        The output generated usually by the network
    target : torch.Tensor
        The label for the corresponding Tensor for which the output was generated
    reduction : string, optional
        DESCRIPTION. The default is 'mean'.
    scaling_factor : integer, optional
        The scaling factor to multiply the label with

    Returns
    -------
    loss : torch.Tensor
        Computed Mean Squared Error loss for the output and label

    """
    scaling_factor = torch.as_tensor(scaling_factor)
    label = label.float()
    label = label * scaling_factor
    loss_fn = MSELoss(reduction=reduction)
    iflat = output.contiguous().view(-1)
    tflat = label.contiguous().view(-1)
    loss = loss_fn(iflat, tflat)
    return loss


def MSE_loss(inp, target, params):
    acc_mse_loss = 0
    # if inp.shape != target.shape:
    #     sys.exit('Input and target shapes are inconsistent')

    reduction = "mean"
    if params is not None:
        if "mse" in params["loss_function"]:
            if isinstance(params["loss_function"]["mse"], dict):
                reduction = params["loss_function"]["mse"]["reduction"]

    if inp.shape[0] == 1:
        if params is not None:
            acc_mse_loss += MSE(
                inp,
                target,
                reduction=reduction,
                scaling_factor=params["scaling_factor"],
            )
        else:
            acc_mse_loss += MSE(inp, target)
        # for i in range(0, params["model"]["num_classes"]):
        #    acc_mse_loss += MSE(inp[i], target[i], reduction=params["loss_function"]['mse']["reduction"])
    else:
        if params is not None:
            for i in range(0, params["model"]["num_classes"]):
                acc_mse_loss += MSE(
                    inp[:, i, ...],
                    target[:, i, ...],
                    reduction=reduction,
                    scaling_factor=params["scaling_factor"],
                )
        else:
            for i in range(0, inp.shape[1]):
                acc_mse_loss += MSE(inp[:, i, ...], target[:, i, ...])
    if params is not None:
        acc_mse_loss /= params["model"]["num_classes"]
    else:
        acc_mse_loss /= inp.shape[1]

    return acc_mse_loss



# -*- coding: utf-8 -*-
"""
Implementation of MSDNet
"""

import torch.nn.functional as F
import torch.nn as nn
import torch
from GANDLF.models.seg_modules.add_conv_block import add_conv_block
from .modelBase import ModelBase


class MSDNet(ModelBase):
    """
    Paper: A mixed-scale dense convolutional neural network for image analysis
    Published: PNAS, Jan. 2018
    Paper: http://www.pnas.org/content/early/2017/12/21/1715832114
    DOI: 10.1073/pnas.1715832114
    Derived from Shubham Dokania's https://github.com/shubham1810/MS-D_Net_PyTorch
    """

    @staticmethod
    def weight_init(m):
        if isinstance(m, nn.Linear):
            nn.init.kaiming_normal_(m, m.weight.data)

    def __init__(
        self,
        parameters: dict,
        num_layers=4,
    ):
        super(MSDNet, self).__init__(parameters)

        self.layer_list = add_conv_block(
            self.Conv, self.BatchNorm, in_ch=self.n_channels
        )

        current_in_channels = 1
        # Add N layers
        for i in range(num_layers):
            s = i % 10 + 1
            self.layer_list += add_conv_block(
                self.Conv, self.BatchNorm, in_ch=current_in_channels, dilate=s
            )
            current_in_channels += 1

        # Add final output block
        self.layer_list += add_conv_block(
            self.Conv,
            self.BatchNorm,
            in_ch=current_in_channels + self.n_channels,
            out_ch=self.n_classes,
            kernel_size=1,
            last=True,
        )

        # Add to Module List
        self.layers = nn.ModuleList(self.layer_list)

        self.apply(self.weight_init)

    def forward(self, x):
        prev_features = []
        inp = x

        for i, f in enumerate(self.layers):
            # Check if last conv block
            if i == len(self.layers) - 2:
                x = torch.cat(prev_features + [inp], 1)

            x = f(x)

            if (i + 1) % 2 == 0 and not i == (len(self.layers) - 1):
                x = F.relu(x)
                # Append output into previous features
                prev_features.append(x)
                x = torch.cat(prev_features, 1)
        return x



# -*- coding: utf-8 -*-
# adapted from https://github.com/qubvel/segmentation_models.pytorch
from typing import Optional, Union, List
import torch

from segmentation_models_pytorch.base import (
    SegmentationHead,
    ClassificationHead,
)
from segmentation_models_pytorch.encoders import get_encoder
from segmentation_models_pytorch.decoders.unet.decoder import UnetDecoder
from segmentation_models_pytorch.base import initialization as init

from .modelBase import ModelBase


class SegmentationModel(torch.nn.Module):
    """
    This has been adapted from its original implementation in
    https://github.com/qubvel/segmentation_models.pytorch/blob/master/segmentation_models_pytorch/base/model.py

    """

    def initialize(self):
        if self.classification_head is None:
            init.initialize_decoder(self.decoder)
            init.initialize_head(self.segmentation_head)
        else:
            init.initialize_head(self.classification_head)

    def forward(self, x):
        """Sequentially pass `x` trough model`s encoder, decoder and heads"""
        features = self.encoder(x)

        if self.classification_head is None:
            decoder_output = self.decoder(*features)
            masks = self.segmentation_head(decoder_output)
            return masks
        else:
            labels = self.classification_head(features[-1])
            return labels

    ## commented out because we are not using this interface
    # def predict(self, x):
    #     """Inference method. Switch model to `eval` mode, call `.forward(x)` with `torch.no_grad()`

    #     Args:
    #         x: 4D torch tensor with shape (batch_size, channels, height, width)

    #     Return:
    #         prediction: 4D torch tensor with shape (batch_size, classes, height, width)

    #     """
    #     if self.training:
    #         self.eval()

    #     with torch.no_grad():
    #         x = self.forward(x)

    #     return x


class Unet(SegmentationModel):
    """
    This has been adapted from its original implementation in
    https://github.com/qubvel/segmentation_models.pytorch/blob/master/segmentation_models_pytorch/decoders/unet/model.py

    Unet_ is a fully convolution neural network for image semantic segmentation. Consist of *encoder*
    and *decoder* parts connected with *skip connections*. Encoder extract features of different spatial
    resolution (skip connections) which are used by decoder to define accurate segmentation mask. Use *concatenation*
    for fusing decoder blocks with skip connections.

    Args:
        encoder_name: Name of the classification model that will be used as an encoder (a.k.a backbone)
            to extract features of different spatial resolution
        encoder_depth: A number of stages used in encoder in range [3, 5]. Each stage generate features
            two times smaller in spatial dimensions than previous one (e.g. for depth 0 we will have features
            with shapes [(N, C, H, W),], for depth 1 - [(N, C, H, W), (N, C, H // 2, W // 2)] and so on).
            Default is 5
        encoder_weights: One of **None** (random initialization), **"imagenet"** (pre-training on ImageNet) and
            other pretrained weights (see table with available weights for each encoder_name)
        decoder_channels: List of integers which specify **in_channels** parameter for convolutions used in decoder.
            Length of the list should be the same as **encoder_depth**
        decoder_use_batchnorm: If **True**, BatchNorm2d layer between Conv2D and Activation layers
            is used. If **"inplace"** InplaceABN will be used, allows to decrease memory consumption.
            Available options are **True, False, "inplace"**
        decoder_attention_type: Attention module used in decoder of the model. Available options are **None** and **scse**.
            SCSE paper - https://arxiv.org/abs/1808.08127
        in_channels: A number of input channels for the model, default is 3 (RGB images)
        classes: A number of classes for output mask (or you can think as a number of channels of output mask)
        activation: An activation function to apply after the final convolution layer.
            Available options are **"sigmoid"**, **"softmax"**, **"logsoftmax"**, **"tanh"**, **"identity"**, **callable** and **None**.
            Default is **None**
        aux_params: Dictionary with parameters of the auxiliary output (classification head). Auxiliary output is build
            on top of encoder if **aux_params** is not **None** (default). Supported params:
                - classes (int): A number of classes
                - pooling (str): One of "max", "avg". Default is "avg"
                - dropout (float): Dropout factor in [0, 1)
                - activation (str): An activation function to apply "sigmoid"/"softmax" (could be **None** to return logits)

    Returns:
        ``torch.nn.Module``: Unet

    .. _Unet:
        https://arxiv.org/abs/1505.04597

    """

    def __init__(
        self,
        encoder_name: str = "resnet34",
        encoder_depth: int = 5,
        encoder_weights: Optional[str] = "imagenet",
        decoder_use_batchnorm: bool = True,
        decoder_channels: List[int] = (256, 128, 64, 32, 16),
        decoder_attention_type: Optional[str] = None,
        in_channels: int = 3,
        classes: int = 1,
        activation: Optional[Union[str, callable]] = None,
        aux_params: Optional[dict] = None,
    ):
        super().__init__()

        self.encoder = get_encoder(
            encoder_name,
            in_channels=in_channels,
            depth=encoder_depth,
            weights=encoder_weights,
        )

        if aux_params is None:
            self.decoder = UnetDecoder(
                encoder_channels=self.encoder.out_channels,
                decoder_channels=decoder_channels,
                n_blocks=encoder_depth,
                use_batchnorm=decoder_use_batchnorm,
                center=True if encoder_name.startswith("vgg") else False,
                attention_type=decoder_attention_type,
            )
            self.segmentation_head = SegmentationHead(
                in_channels=decoder_channels[-1],
                out_channels=classes,
                activation=activation,
                kernel_size=3,
            )
            self.classification_head = None
        else:
            self.classification_head = ClassificationHead(
                in_channels=self.encoder.out_channels[-1], **aux_params
            )
            self.segmentation_head = None

        self.name = "u-{}".format(encoder_name)
        self.initialize()


class ImageNet_UNet(ModelBase):
    """
    Wrapper class to parse arguments from configuration and construct appropriate model object.

    Args:
        ModelBase (nn.Module): The base model class.
    """

    def __init__(
        self,
        parameters,
    ) -> None:
        super(ImageNet_UNet, self).__init__(parameters)

        decoder_use_batchnorm = False
        if parameters["model"]["norm_type"] == "batch":
            decoder_use_batchnorm = True
        decoder_use_batchnorm = parameters["model"].get(
            "decoder_use_batchnorm", decoder_use_batchnorm
        )
        encoder_depth = parameters["model"].get("depth", 5)
        encoder_depth = parameters["model"].get("encoder_depth", encoder_depth)
        parameters["model"]["pretrained"] = parameters["model"].get("pretrained", True)
        if parameters["model"]["pretrained"]:
            parameters["model"]["encoder_weights"] = "imagenet"
        else:
            parameters["model"]["encoder_weights"] = None

        if parameters["problem_type"] != "segmentation":
            classifier_head_parameters = {}
            classifier_head_parameters["classes"] = self.n_classes
            classifier_head_parameters["activation"] = parameters["model"][
                "final_layer"
            ]
            if classifier_head_parameters["activation"] == "None":
                classifier_head_parameters["activation"] = None
            classifier_head_parameters["dropout"] = parameters["model"].get(
                "dropout", 0.2
            )
            classifier_head_parameters["pooling"] = parameters["model"].get(
                "pooling", "avg"
            )
        else:
            classifier_head_parameters = None

        self.model = Unet(
            encoder_name=parameters["model"].get("encoder_name", "resnet34"),
            encoder_weights=parameters["model"]["encoder_weights"],
            in_channels=self.n_channels,
            classes=self.n_classes,
            activation=parameters["model"]["final_layer"],
            decoder_use_batchnorm=decoder_use_batchnorm,
            decoder_attention_type=parameters["model"].get(
                "decoder_attention_type", None
            ),
            encoder_depth=encoder_depth,
            decoder_channels=parameters["model"].get(
                "decoder_channels", (256, 128, 64, 32, 16)
            ),
            aux_params=classifier_head_parameters,
        )

        if self.n_dimensions == 3:
            self.model = self.converter(self.model).model

    def forward(self, x):
        return self.model(x)


def imagenet_unet_wrapper(parameters):
    return ImageNet_UNet(parameters)



import torch.nn as nn


class IncDropout(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        Conv,
        Dropout,
        InstanceNorm,
        dropout_p=0.3,
        leakiness=1e-2,
        conv_bias=True,
        inst_norm_affine=True,
        res=False,
        lrelu_inplace=True,
    ):
        nn.Module.__init__(self)
        self.input_channels = input_channels
        self.output_channels = output_channels
        self.dropout_p = dropout_p
        self.leakiness = leakiness
        self.conv_bias = conv_bias
        self.inst_norm_affine = inst_norm_affine
        self.residual = res
        self.lrelu_inplace = lrelu_inplace
        self.dropout = Dropout(dropout_p)
        self.conv = Conv(
            input_channels,
            output_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=self.conv_bias,
        )

    def forward(self, x):
        x = self.dropout(x)
        x = self.conv(x)
        return x
